#include "../ansi/r.h"
#include "../ansi/mynr.h"
#include "../ansi/macopt.h"

/* 
   infera
   this program uses macopt to solve for the most probable team potentials
   given a file of scores

   each team t=1..T has a potential a[t] such that when two teams are 
   playing each other, the probability that team t1 gets the next point
   is exp(a[t1])/(exp(a[t1])+exp(a[t2]))

   The program is described in http://wol.ra.phy.cam.ac.uk/ultimate/infera/

   usage

   infera -cg 1 -t 2 -g 1 -w 0 -ov 2 -s score2 
   infera -cg 1 -t 3 -g 3 -w 0 -ov 2 -s score3 
   infera -cg 1 -cge 0.00001 -t 3 -g 3 -w 0 -ov 1 -v 1 -s score3 
   infera -spread 20 -cg 1 -cge 0.000001 -t 3 -g 3 -w 0 -ov 1 -v 1 -s score3 
   infera -t 3 -g 3 -ov 1 -v 1 -s score3 
   infera -t 3 -g 3 -ov 0 -tol 0.0001 -v 1 -s score3 
*/
   
#include "scores.h"
static int read_tms ( opt_args *  ) ;
static int make_space ( opt_args *  ) ;
static int free_space ( opt_args *  ) ; 
double logz ( double , double  ) ;
static int read_scores ( opt_args *  ) ;
static void print_usage ( char ** , FILE * ,
			 opt_args * );
static int process_command ( int  , char ** , opt_args  * ) ;
static int initial_condn ( opt_args *  ) ;
static void print_state ( opt_args * , FILE * ) ;
static void diagnose ( opt_args * , FILE * ) ;
static void defaults ( opt_args * )  ;
static void myopt_defaults ( macopt_args * , opt_args * )  ;
void grad_logprob(double * , double *    , void * ) ;
double logprob(double * ,   void * ) ;
void main(int , char ** );

void main(int argc, char *argv[])
{
  opt_args ss ; /* the scores */
  macopt_args a ;
  FILE * fp ; 
  

 defaults   ( &ss ) ; 
  if ( process_command (argc, argv, &ss ) < 0 ) exit (0) ;
  if (  make_space ( &ss ) < 0 ) exit (0) ;
  if (  read_scores ( &ss ) < 0 ) exit (0) ;
  if (  ss.readt && read_tms ( &ss ) < 0 ) exit (0) ;

  /* Load up the parameters of the function that you want to optimize */
  initial_condn ( &ss ) ; 
  /* any other things to set up? */
  ss.priorfactor = 1.0 / (2.0*ss.spread*ss.spread) ;

  /* Check that the gradient_function is the gradient of the function  */
  /* You don't have to do this, but it is a good idea when debugging ! */
  if ( ss.cg ) maccheckgrad (  ss.a , ss.T , ss.cgepsilon , 
		logprob , (void *)(&ss) , 
		grad_logprob , (void *)(&ss) , 
		0
		) ;

  /* initialize the arguments of the optimizer */
  macopt_defaults ( &a ) ; 
  /* modify them from default values */
  myopt_defaults ( &a , &ss ) ;

  /* Do an optimization */
  macoptII ( ss.a , ss.T , 
	    grad_logprob , (void *)(&ss) , &a
	    ) ;

  if ( ss.verbose >= 1 ) {
    printf("Solution:\n");
    print_state ( &ss , stdout ) ;
    logprob( ss.a ,&ss);
  }
  if ( ss.out) {
    fp = fopen ( ss.ofile , "w" ) ;
    if (!fp ) fprintf ( stderr , "can't open file %s \n" , ss.ofile ) ;
    else {
      print_state ( &ss , fp ) ; 
      fclose ( fp ) ; 
    }
  }
  if ( ss.diagnostics ) diagnose ( &ss , stdout ) ; 
  if ( ss.cg ) maccheckgrad (  ss.a , ss.T , ss.cgepsilon , 
		logprob , (void *)(&ss) , 
		grad_logprob , (void *)(&ss) , 
		0
		) ;

  if (  free_space ( &ss ) < 0 ) exit (0) ;

}

static void myopt_defaults ( macopt_args *a , opt_args *s  ) {
  a->tol = s->tol ; 
  a->end_if_small_step = s->endifsmall ; 
  a->itmax = s->itmax ;       
  a->rich = s->rich ;  
  a->verbose = s->optverbose ; 
}

static int initial_condn ( opt_args *ss  ) {
  int status = 0 ; 
  int t ; 

  for ( t = 1 ; t<= ss->T ; t ++ ) {
    ss->a[t] = ss->a0 ; 
  }
  return status ; 
}

static int make_space ( opt_args *ss  ) {
  int status = 0 ; 

  ss->a = dvector ( 1 , ss->T ) ; 
  ss->w = dvector ( 1 , ss->G ) ; 
  ss->lz = dvector ( 1 , ss->G ) ; 
  ss->s = imatrix ( 1 , ss->G , 1 , 4 ) ; /* array of scores */

  return status ; 
}

static int free_space ( opt_args *s  ) {
  int status = 0 ; 

 free_dvector (  s->a , 1 , s->T ) ; 
  free_dvector ( s->w , 1 , s->G ) ; 
  free_dvector ( s->lz , 1 , s->G ) ; 
  free_imatrix ( s->s,  1 , s->G , 1 , 4 ) ; /* array of scores */

  return status ; 
}

static int read_tms ( opt_args *ss  ) {
  FILE    *fp;
  char *file = ss->tfile ;
  int t , t1 , status = 0 ; 
  fp = fopen( file, "r" );
  if( !fp )   fprintf( stderr, "No such file: %s\n", file ), exit(0);
  printf ( "Reading teams from %s... ", file ) ; fflush ( stdout ) ; 
  for ( t = 1 ; t <= ss->T ; t ++ ) {
    if ( fscanf(fp,"%d %s" , &t1 , ss->tms[t] ) == EOF ) {
      status = -1 ;       break ;
    }

  }
  if ( status < 0 ) { 
    fprintf ( stderr , "error reading line %d of file %s\n" , 
			       t , file ) ; 
  }
  printf ( "done\n" ) ;

  return status ; 
}


static int read_scores ( opt_args *ss  ) {
  /* reads in each line of the file */
  /* one error-handling strategy would be:
     if any team appears twice on one row
     or is outside 1,T then generate warning, 
     ditch that row, and reduce number of games 

     the current version crashes on error */

  int status = 0 ; 
  FILE    *fp;
  char *file = ss->sfile ;
  int g , t1 , s1 , t2 , s2; 
  double *w = ss->w ; 
  int **s = ss->s ; 
  
  fp = fopen( file, "r" );
  if( !fp )   fprintf( stderr, "No such file: %s\n", file ), exit(0);
  printf ( "Reading scores from %s... ", file ) ; fflush ( stdout ) ; 

  for ( g = 1 ; g <= ss->G ; g ++ ) {
    if ( fscanf(fp,"%d %d %d %d " , &t1 , &s1 , &t2 , &s2 ) == EOF ) {
      status = -1 ;       break ;
    }
    if ( ss->weighted ) {
      if ( fscanf(fp,"%lf " , &w[g] ) == EOF ) {
	status = -1 ;       break ;}
    }
    s[g][1] = t1 ; s[g][2] = s1 ; s[g][3] = t2 ; s[g][4] = s2 ; 
    if ( ( t1 > ss->T ) ||  ( t1 < 1 )  ||  ( t1 > ss->T ) ||  ( t1 < 1 ) ) {
      fprintf ( stderr , "team number out of range 1,%d\n", ss->T ) ; 
      status -- ; 
    }
    if ( ( t1 == t2 ) ) {
      fprintf ( stderr , "team playing itself? %d\n", t1 ) ; 
      status -- ; 
    }
    if ( s1 < 0 || s2 < 0 ) {
      fprintf ( stderr , "bizarre scores %d %d\n", s1 , s2 ) ; 
      status -- ; 
    }
  }
  if ( status < 0 ) { 
    fprintf ( stderr , "error reading line %d of file %s\n" , 
			       g , file ) ; 
  }
  printf ( "done\n" ) ;

  return status ; 
}

static void print_state ( opt_args *ss , FILE *fp ) {
  int i ; 
  for ( i = 1 ; i <= ss->T ; i++ ) {
    fprintf( fp , "%3d: " , i ) ;
    if ( ss->readt ) fprintf( fp , "%10s " , ss->tms[i] ) ;
    fprintf( fp , "%10.3f %10.1f\n" , ss->a[i],  ss->reportfact * ss->a[i] + ss->reportmean ) ;
  }
} 

static void diagnose ( opt_args *ss , FILE *fp ) { /* checks each game to see 
						      how big a discrepancy */
  int  g,t1,t2 ;
  double /* f=0.0 ,*/ w, a1, a2 ,s1,s2 , stot, p1 , p2 , gr1 , gr2 ;
  int **s = ss->s ; 
  double *a = ss->a  ; 
  double *lz = ss->lz ; 

  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; a1=a[t1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; a2=a[t2] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      s1 *= w ; 
      s2 *= w ; 
    }
    lz[g] = logz ( a1 , a2 ) ; 
    stot = s1+s2 ; 
    if ( a1 < a2 ) { p1 = exp ( a1 -  lz[g] ) ; p2 = 1.0 - p1 ; }
    else { p2 = exp ( a2 -  lz[g] ) ; p1 = 1.0 - p2 ; }
/*       f += s1 * a1 + s2 * a2 - ( s1 + s2 ) * lz[g] ; */
/* so    df/da1 += s1 - (s1+s2) * p1                    */
    gr1 = s1 - stot * p1 ; /* minus because f -> -f */
    gr2 = s2 - stot * p2 ; 

    fprintf( fp , "%3d: %8.2g %8.2g " , g , gr1 , gr2 ) ;
    if ( ss->readt ) fprintf( fp , "%10s " , ss->tms[t1] ) ;
    else  fprintf( fp , "%2d " , t1 ) ;
    fprintf( fp , "%2.0f %6.2f  " , s1 ,  stot * p1 ) ;
    if ( ss->readt ) fprintf( fp , "%10s " , ss->tms[t2] ) ;
    else  fprintf( fp , "%2d " , t2 ) ;
    fprintf( fp , "%2.0f %6.1f\n" , s2 ,  stot * p2 ) ;

    /* things to scatter plot */
  }
} 

double logprob(double *x,   void *arg )
{
  int i , g,t1,t2 ;
  double f=0.0 , w, a1, a2 ,s1,s2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *a = x ; 
  double *lz = ss->lz ; 
 
  if ( ss->verbose >= 2 ) 
    {  printf ( "log prob. at " ) ;
       for ( i = 1 ; i <= ss->T ; i++ ) printf( "%g " , x[i] ) ;
     } 
  ss->logprior = 0.0 ; 
  for ( i = 1 ; i <= ss->T ; i++ ) ss->logprior -= x[i]*x[i]* ss->priorfactor ;
  /* run through games */
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; a1=a[t1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; a2=a[t2] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      s1 *= w ; 
      s2 *= w ; 
    }
    lz[g] = logz ( a1 , a2 ) ; 
    f += s1 * a1 + s2 * a2 - ( s1 + s2 ) * lz[g] ;
  }
  ss->loglike = f ; 
  f += ss->logprior ; 
  if ( ss->verbose >= 2 ) 
    printf(" : %g\n",f);
  else if ( ss->verbose >= 1 ) 
    printf("logprob : %g\n",f);
  return(-f); 
}

double logz ( double a1 , double a2 ) {
  double lz , ama , ami ; 
  if ( a1 > a2 ) { ama = a1 ; ami = a2 ; }
  else { ama = a2 ; ami = a1 ; }
/*  lz = log (  exp(ama) + exp(ami) ) 
       = ama + log ( 1 + exp(ami-ama) ) */
  lz = ama + log ( 1.0 + exp ( ami - ama ) ) ;
  return lz ; 
}

void grad_logprob(double *x, double *gr, void *arg )
{
  int i , g,t1,t2 ;
  double /* f=0.0 ,*/ w, a1, a2 ,s1,s2 , stot, p1 , p2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *a = x  ; 
  double *lz = ss->lz ; 

  if ( ss->verbose >= 2 ) {
    printf ( "gradient at " ) ;
    for ( i = 1 ; i <= ss->T ; i++ ) printf ( "%g " , x[i] ) ;
  }
  for ( i = 1 ; i <= ss->T ; i++ ) gr[i] = 2.0 * x[i] * ss->priorfactor ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; a1=a[t1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; a2=a[t2] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      s1 *= w ; 
      s2 *= w ; 
    }
    lz[g] = logz ( a1 , a2 ) ; 
    stot = s1+s2 ; 
    if ( a1 < a2 ) { p1 = exp ( a1 -  lz[g] ) ; p2 = 1.0 - p1 ; }
    else { p2 = exp ( a2 -  lz[g] ) ; p1 = 1.0 - p2 ; }
/*       f += s1 * a1 + s2 * a2 - ( s1 + s2 ) * lz[g] ; */
/* so    df/da1 += s1 - (s1+s2) * p1                    */
    gr[t1] -= s1 - stot * p1 ; /* minus because f -> -f */
    gr[t2] -= s2 - stot * p2 ; 
  }
  if ( ss->verbose >= 2 ) {
    for ( i = 1 ; i <= ss->T ; i++ ) printf ( "%g " , gr[i] ) ;
    printf ( "\n" ) ;
  }
}

/************************ boring stuff *********************************/

static void defaults ( opt_args *s ) 
{
#include "_var_def.c"
}

static int process_command ( int argc , char **argv , opt_args *s ) {

  int p_usage = 0 ;
  int status = 0 ;
  int cs , i ;

  if ( argc < 1 )     {
    p_usage = 1 ; 
    status -- ;
  }

#define ERROR1 fprintf ( stderr , "arg to `%s' missing\n" , argv[i] ) ; \
               status --
#define ERROR2 fprintf ( stderr , "args to `%s' missing\n" , argv[i] ) ; \
               status --
#define ERRORREG fprintf ( stderr , "regtype must be defined before `%s'\n" , argv[i] ) ; \
               status --
  for (i = 1 ; i < argc; i++)    {
    cs = 1 ;
    if ( strcmp (argv[i], "-V") == 0 )        {
      s->verbose = 1;
    }
#include "_var_clr.c"
    else {
      fprintf ( stderr , "arg `%s' not recognised\n" , argv[i] ) ; 
      p_usage = 1 ;
      status -- ;
    }
    if ( cs == 0 ) {
      fprintf ( stderr , "arg at or before `%s' has incorrect format\n" , 
	       argv[i] ) ;
      p_usage = 1 ;
      status -- ;
    }
  }
  if ( p_usage ) print_usage ( argv , stderr , s ) ;
  return ( status ) ;
}
#undef ERROR1
#undef ERROR2
#undef ERRORREG

#define DNT fprintf( fp, "\n        ")
#define NLNE  fprintf( fp, "\n")

static void print_usage ( char **argv , FILE * fp , opt_args *s )
{
  fprintf( fp, "Usage: %s ",argv[0]);
  fprintf( fp, " [optional arguments]");
  DNT;  fprintf( fp, "-V | -VV                 (verbose or very verbose)"); 
  NLNE;
#include "_var_usg.c"
  NLNE; 
  fprintf( fp, "\n");
  return ;
}
