#include "../ansi/r.h"
#include "../ansi/mynr.h"
#include "../ansi/macopt.h"

/* 
   infera 
   this program uses macopt to solve for the most probable team potentials
   given a file of scores

   type 1:

   each team t=1..T has a potential a[t] such that when two teams are 
   playing each other, the probability that team t1 gets the next point
   is exp(a[t1])/(exp(a[t1])+exp(a[t2]))

   type 2: 

   assumes an o ability and d ability (for games with fixed duration only)

   type 4: assumes a home - away effect too. first colunm is home team.

   type 5: assumes poisson rates in each game between two teams may differ 
   with fluctuations described by parameters "c" (Gamma distribution). (Positive correlation in flucns)

   type 6: assumes poisson rates in each game between two teams may differ 
   with fluctuations described by parameters "c" (Gamma distribution) (No correlation in flucns)


   The program is described in http://www.inference.phy.cam.ac.uk/ultimate/infera/

   usage

   infera -cg 1 -t 2 -g 1 -w 0 -ov 2 -s score2 
   infera -cg 1 -t 3 -g 3 -w 0 -ov 2 -s score3 
   infera -cg 1 -cge 0.00001 -t 3 -g 3 -w 0 -ov 1 -v 1 -s score3 
   infera -spread 20 -cg 1 -cge 0.000001 -t 3 -g 3 -w 0 -ov 1 -v 1 -s score3 
   infera -t 3 -g 3 -ov 1 -v 1 -s score3 
   infera -t 3 -g 3 -ov 0 -tol 0.0001 -v 1 -s score3 
*/
   
#include "scores.h"
static int read_tms ( opt_args *  ) ;
static int make_space ( opt_args *  ) ;
static int free_space ( opt_args *  ) ; 
double logz ( double , double  ) ;
static int read_scores ( opt_args *  ) ;
static void print_usage ( char ** , FILE * ,
			 opt_args * );
static int process_command ( int  , char ** , opt_args  * ) ;
static int initial_condn ( opt_args *  ) ;
static void final_set_up ( opt_args *  ) ;
static void reestimate ( opt_args *s ) ;
static void do_table ( opt_args *s , FILE *fp ) ;
static void print_state ( opt_args * , FILE * ) ;
static void diagnose ( opt_args * , FILE * ) ;
static void wld ( double *p1 , double *p2 , double *win , double *lose , 
		 double *draw , double *sdraw , int maxg ) ;
static void print_odds ( opt_args *ss , FILE *fp , FILE *fpl ) ;
static double odds ( double p , opt_args *s ) ;
static void odds_print ( FILE * fp , double o ) ;
static int poisson ( double l , double *p , int maxg ) ;
static void print_odd_head ( FILE *fp ) ;
static void  pg_process ( opt_args *s ) ;
void marginal_p ( double *p , double lam , double sdl , double c , double sdc , double *scratch , int maxg , double *logfact ) ;
static void  ugly_print_score_table  ( opt_args *s , FILE *fp  ) ;
static void  pretty_print_score_table  ( opt_args *s , FILE *fp  ) ;
static void increment_poisson_product ( double **pg , double *p1 , 
                 double *p2 , double *win , double *lose , 
		 double *draw, double *sdraw , int maxg , double w , 
				       double * , int * , int * ) ;
static void increment_gamma_product ( opt_args *s , double **pg , double llc , double lc ,
				     double c , double ll1 , double ll2 , 
 double *win , double *lose , 
		 double *draw, double *sdraw , int maxg , double weight,
	       double *pmps , int *mpg1 , int *mpg2  ) ;
static void defaults ( opt_args * )  ;
static void myopt_defaults ( macopt_args * , opt_args * )  ;
void grad_logprob(double * , double *    , void * ) ;
double logprob(double * ,   void * ) ;
void find_A ( opt_args *ss ) ;
void grad_logprob2(double * , double *    , void * ) ;
double logprob2(double * ,   void * ) ;
void find_A2 ( opt_args *ss ) ;
void grad_logprob4(double * , double *    , void * ) ;
double logprob4(double * ,   void * ) ;
void find_A4 ( opt_args *ss ) ;
void grad_logprob5(double * , double *    , void * ) ;
double logprob5(double * ,   void * ) ;
void find_A5 ( opt_args *ss ) ;
void grad_logprob6(double * , double *    , void * ) ;
double logprob6(double * ,   void * ) ;
void find_A6 ( opt_args *ss ) ;
void main(int , char ** );

void main(int argc, char *argv[])
{
  opt_args ss ; /* the scores */
  macopt_args a ;
  FILE * fp , *fpl ; 
  double det; 

 defaults   ( &ss ) ; 
  if ( process_command (argc, argv, &ss ) < 0 ) exit (0) ;
  if (  make_space ( &ss ) < 0 ) exit (0) ;
  if (  read_scores ( &ss ) < 0 ) exit (0) ;
  if (  ss.readt && read_tms ( &ss ) < 0 ) exit (0) ;

  /* Load up the parameters of the function that you want to optimize */
  initial_condn ( &ss ) ; 
  /* any other things to set up? */
  final_set_up ( &ss ) ; 

  macopt_defaults ( &a ) ;    /* initialize the arguments of the optimizer */
  myopt_defaults ( &a , &ss ) ;   /* modify them from default values */

  if ( ss.type == 1 ) { /* the original version with each team having an "a" */
    /* Check that the gradient_function is the gradient of the function  */
    if ( ss.cg ) maccheckgrad (  ss.a , ss.k , ss.cgepsilon , 
			       logprob , (void *)(&ss) , 
			       grad_logprob , (void *)(&ss) , 0) ;
    macoptII ( ss.a , ss.k ,  grad_logprob , (void *)(&ss) , &a ) ;

    if ( ss.find_ebs ) {
      find_A ( &ss ) ; 
      det = invert_dmatrix_family ( &(ss.A) , 1 ) ;
      if ( ss.make_eb_table ) {
	do_table ( &(ss) , stdout ) ;
      }
    }
    if ( ss.diagnostics ) diagnose ( &ss , stdout ) ; 
    if ( ss.cg ) maccheckgrad (  ss.a , ss.k , ss.cgepsilon , 
			       logprob , (void *)(&ss) , grad_logprob , (void *)(&ss) , 0 ) ;
  } /* end original type 1 version */
  else if ( ss.type == 2 ) { /* each team has an ao and an ad */
    if ( ss.cg ) maccheckgrad (  ss.a , ss.k , ss.cgepsilon , 
			       logprob2 , (void *)(&ss) , 
			       grad_logprob2 , (void *)(&ss) , 0) ;
    macoptII ( ss.a , ss.k ,  grad_logprob2 , (void *)(&ss) , &a ) ;

    if ( ss.find_ebs ) {
      find_A2 ( &ss ) ; 
      det = invert_dmatrix_family ( &(ss.A) , 1 ) ;
/*      if ( ss.make_eb_table ) {
	do_table2 ( &(ss) , stdout ) ;
      } */
    }
/*    if ( ss.diagnostics ) diagnose2 ( &ss , stdout ) ; */
    if ( ss.cg ) 
      maccheckgrad (  ss.a , ss.k , ss.cgepsilon , logprob2 , (void *)(&ss) , grad_logprob2 , (void *)(&ss) , 0 ) ;

  }
  else if ( ss.type == 4 ) { /* each team has an ao and an ad at "away" and a ho and a hd at "home" */
    if ( ss.cg ) maccheckgrad (  ss.a , ss.k , ss.cgepsilon , 
			       logprob4 , (void *)(&ss) , 
			       grad_logprob4 , (void *)(&ss) , 0) ;
    macoptII ( ss.a , ss.k ,  grad_logprob4 , (void *)(&ss) , &a ) ;

    if ( ss.find_ebs ) {
      find_A4 ( &ss ) ; 
      det = invert_dmatrix_family ( &(ss.A) , 1 ) ;
/*      if ( ss.make_eb_table ) {
	do_table4 ( &(ss) , stdout ) ;
      } */
    }
/*    if ( ss.diagnostics ) diagnose4 ( &ss , stdout ) ; */
    if ( ss.cg ) 
      maccheckgrad (  ss.a , ss.k , ss.cgepsilon , logprob4 , (void *)(&ss) , grad_logprob4 , (void *)(&ss) , 0 ) ;

  }  else if ( ss.type == 5 ) { /* each team has an ao and an ad at "away" and a ho and a hd at "home" and contributes logc to a variability measure producing POSITIVE correlation in scores */
    if ( ss.cg ) maccheckgrad (  ss.a , ss.k , ss.cgepsilon , 
			       logprob5 , (void *)(&ss) , 
			       grad_logprob5 , (void *)(&ss) , 0) ;
    macoptII ( ss.a , ss.k ,  grad_logprob5 , (void *)(&ss) , &a ) ;

/*    if ( ss.find_ebs ) {
      find_A4 ( &ss ) ; 
      det = invert_dmatrix_family ( &(ss.A) , 1 ) ;
      }*/
    if ( ss.cg ) 
      maccheckgrad (  ss.a , ss.k , ss.cgepsilon , logprob5 , (void *)(&ss) , grad_logprob5 , (void *)(&ss) , 0 ) ;
  }  else if ( ss.type == 6 ) { /* each team has an ao and an ad at "away" and a ho and a hd at "home" and contributes logco / logcd to a variability measure  */
    if ( ss.cg ) maccheckgrad (  ss.a , ss.k , ss.cgepsilon , 
			       logprob6 , (void *)(&ss) , 
			       grad_logprob6 , (void *)(&ss) , 0) ;
    ss.loop = 0 ; 
    do {
      macoptII ( ss.a , ss.k ,  grad_logprob6 , (void *)(&ss) , &a ) ;
      ss.loop ++ ; 
      if ( ss.loop < ss.reest ) { reestimate ( &ss ) ; }
    } while ( ss.loop < ss.reest ) ;

    if ( ss.find_ebs ) {
      find_A6 ( &ss ) ; 
      det = invert_dmatrix_family ( &(ss.A) , 1 ) ;
      if ( ss.showA ) {
	show_dmatrix_family ( &(ss.A) , stdout ) ; 
      }
    }
    if ( ss.cg ) 
      maccheckgrad (  ss.a , ss.k , ss.cgepsilon , logprob6 , (void *)(&ss) , grad_logprob6 , (void *)(&ss) , 0 ) ;
  } else { printf( "!!! invalid type\n" ); }
  if ( ss.verbose >= 1 ) {
    printf("MAP Solution:\n");
    print_state ( &ss , stdout ) ;
/*    logprob( ss.a ,&ss); */
  }
  if ( ss.out) {
    fp = fopen ( ss.ofile , "w" ) ;
    if (!fp ) fprintf ( stderr , "can't open file %s \n" , ss.ofile ) ;
    else {
      print_state ( &ss , fp ) ; 
      fclose ( fp ) ; 
    }
  }
  if ( ss.odds || ss.odl ) {
    if ( ss.type > 1 ) {
      if ( ss.odds) {
	fp = fopen ( ss.oddsfile , "w" ) ;
	if ( !fp ) fprintf ( stderr , "can't open file %s \n" , ss.oddsfile ) ;
      } else fp = 0 ; 
      if ( ss.odl ) {
	fpl = fopen ( ss.odlfile , "w" ) ;
	if (!fpl ) fprintf ( stderr , "can't open file %s \n" , ss.odlfile ) ;
      } else fpl = 0 ; 
      if ( fp || fpl )  {
	print_odds ( &ss , fp , fpl ) ; 
	if ( fp ) fclose ( fp ) ; 
	if ( fpl ) fclose ( fpl ) ; 
      }
    } else {
      fprintf ( stderr , "can't give odds, using type 1 model\n" ) ;
    }
  }
  if (  free_space ( &ss ) < 0 ) exit (0) ;

}

static void reestimate ( opt_args *s ) {
  int i , j , amin , amax , cmin , cmax ;
  double *a = s->a , mean , alpha , count , ms , n_bad , oldalpha ; 
  s->sum_logc = s->ss_logc = s->sum_a = s->ss_a = 0.0 ; 
  if ( s->type == 2 ) {
    amin = 1 ; amax = s->T * 2 ;
  } else if ( s->type == 4 ||  s->type == 5 ||  s->type == 6 ) {
    amin = 1 ; amax = s->T * 4 ;
  } else { amin = amax = 0 ; }
  if ( s->type == 5 ) {
    cmin = amax + 1 ; cmax = amax + s->T ;
  } else  if ( s->type == 6 ) {
    cmin = amax + 1 ; cmax = amax + s->T * 2 ;
  } else { cmin = cmax = 0 ; }

  count = 0.0 ;
  for ( i = amin ; i <= amax ; i ++ ) {
    s->sum_a += a[i] ; 
    s->ss_a += a[i] * a[i] ; 
    count ++ ; 
  }
  if ( count > 0.0 ) {
    oldalpha = s->alpha[amin] ;
    /* number of params that are 'bad' */ /* (double) ( s->ka ) */
    n_bad = count * 
      oldalpha / ( s->aloglikecurvature + oldalpha ) ; 
    if ( n_bad < 1.0 ) { n_bad = 1.0 ; }
    mean = s->sum_a / count ; 
    ms = s->ss_a / count ;
    for ( j = 1 ; j <= s->alpha_its ; j ++ ) {
      alpha = ( count + s->a_alpha_c0 )  / 
	( ( n_bad/oldalpha) + count * ( ms - mean*mean ) 
	 + s->a_alpha_c0 / s->a_alpha_0 ) ;
      oldalpha = alpha ; 
    } 
    for ( i = amin ; i <= amax ; i ++ ) {
      s->mu[i] = mean ; 
      s->alpha[i] = alpha ; 
    }
    printf(" a's mu -> %g , alpha -> %g\n" , mean , alpha ) ; 
  }

  if ( s->type == 5 ||  s->type == 6 ) {
    count = 0.0 ;
    for ( i = cmin ; i <= cmax ; i ++ ) {
      count ++ ;
      s->sum_logc += a[i] ; 
      s->ss_logc += a[i] * a[i] ; 
    }
    if ( count > 0.0 ) {
      oldalpha = s->alpha[cmin] ;
      n_bad = count * 
	oldalpha / ( s->cloglikecurvature + oldalpha ) ; 
      if ( n_bad < 1.0 ) { n_bad = 1.0 ; }
      mean = s->sum_logc / count ; 
      ms = s->ss_logc / count ;
      for ( j = 1 ; j <= s->alpha_its ; j ++ ) {
	alpha = ( count + s->c_alpha_c0 )  / 
	  ( ( n_bad/oldalpha) + count * ( ms - mean*mean ) 
	   + s->c_alpha_c0 / s->c_alpha_0 ) ;
	oldalpha = alpha ; 
      }
/* was:
      alpha = s->logc_fudge / ( ms - mean*mean ) ; */
      for ( i = cmin ; i <= cmax ; i ++ ) {
	s->mu[i] = mean ; 
	s->alpha[i] = alpha ; 
      }
      printf(" logc mu -> %g , alpha -> %g\n" , mean , alpha ) ; 
    }
  }
}

static void do_table ( opt_args *s , FILE *fp ) {
  int i , j , k , t1 , t2 ;
  double *v1, *v2 , var ;  /* type 1 only */
  /* makes a table of values of the variance of the quanity
     a1-a2 for all teams by evaluating g A^{-1} g 
     for g = ( 00000000 -1 0000 1 000 )
     */
  if ( s->type != 1 ) { fprintf ( stderr , "warning, this routine for type 1 only \n" ) ; }
  v1 = dvector ( 1 , s->T ) ;
  v2 = dvector ( 1 , s->T ) ;
/*  first go through games seeing which teams met */
  for ( i = 1 ; i <= s->T ; i ++ ) {
    for ( j = 1 ; j <=s->T ; j ++ ) {
      s->meetings[i][j] = 0 ; 
    }
  }
  for ( i = 1 ; i <= s->G ; i ++ ) {
    t1 = s->s[i][1] ;
    t2 = s->s[i][3] ;
    s->meetings[ t1 ][ t2 ] ++ ;
    s->meetings[ t2 ][ t1 ] ++ ;
  }

  for ( i = 1 ; i <= s->T ; i ++ ) {
    fprintf ( fp , "%2d " , i ) ; 
    for ( j = 1 ; j < i ; j ++ ) {
      for ( k = 1 ; k <= s->T ; k ++ ) {
	v1[k] = v2[k] = 0.0 ;
      } 
      v1[i] = v2[i] = 1.0 ; 
      v1[j] = v2[j] = -1.0 ; 
      var =  lu_quadratic_form ( v1, &(s->A) , v2 ) ;
      if ( s->meetings[ i][j] ) fprintf ( fp , "%d" , s->meetings[ i][j] ) ; 
      else fprintf ( fp , " " ) ; 
      fprintf ( fp , " %4.2g " ,  sqrt(var) ) ; 
    }
    fprintf ( fp , "\n" ) ; 
  }
}

static void final_set_up ( opt_args *s  ) {
  int i ; 
  s->aloglikecurvature = s->a_fudge * (double)(s->total_scores) / ( double)(s->ka ) ; 
  s->cloglikecurvature = s->logc_fudge * (double)(s->total_scores) / ( double)(s->kc ) ; 
  s->priorfactor = 1.0 / (2.0*s->spread*s->spread) ;
  if ( s->type == 5 || s->type == 6 ) {
    for ( i = 1 ; i <= s->T * 4 ; i ++ ) {
      s->mu[i] = 0.0 ; 
      s->alpha[i] = s->priorfactor * 2.0 ; 
    }
    for (  ; i <= s->k ; i ++ ) {
      s->mu[i] = s->logc_mu ; 
      s->alpha[i] = s->logc_alpha ; 
    }
  }
}

static void myopt_defaults ( macopt_args *a , opt_args *s  ) {
  a->tol = s->tol ; 
  a->end_if_small_step = s->endifsmall ; 
  a->itmax = s->itmax ;       
  a->rich = s->rich ;  
  a->verbose = s->optverbose ; 
}

static int initial_condn ( opt_args *ss  ) {
  int status = 0 ; 
  int t ; 

  if ( ss->type == 5 || ss->type == 6 ) {
    for ( t = 1 ; t<= ss->k ; t ++ ) {
      ss->a[t] = ss->mu[t] ; 
    }
  } else {
    for ( t = 1 ; t<= ss->k ; t ++ ) {
      ss->a[t] = ss->a0 ; 
    }
  }
  return status ; 
}

static int make_space ( opt_args *ss  ) {
  int status = 0 ; 
  int g1 , maxg = ss->maxg ; 
  double this ; 

  ss->logfact = dvector ( 0 , maxg ) ; 
  for ( g1 = 0 ,this = 0.0 ; g1 <= maxg ;  ) {
    ss->logfact[g1] = this ; 
    g1 ++ ;    this += log((double)(g1)) ; 
  }
/* number of parameters */
  if ( ss->type == 2 ) {  ss->ka = 2 * ss->T ; ss->kc = 0 ;  }
  else if ( ss->type == 4 ) {  ss->ka = 4 * ss->T ; ss->kc = 0 ;  }
  else if ( ss->type == 5 ) {  ss->ka = 4 * ss->T ; ss->kc = ss->T ;  } 
  else if ( ss->type == 6 ) {  ss->ka = 4 * ss->T ; ss->kc = 2*ss->T ;  } 
  else { ss->ka = ss->T ; ss->kc = 0 ; }
  ss->k = ss->ka + ss->kc ; 
  /* for the "a" model, the number of parameters is one per team */
  ss->a = dvector ( 1 , ss->k ) ; 
  if ( ss->type == 2 || ss->type==4 || ss->type == 5 || ss->type == 6 ) { 
    ss->ao = ss->a ; ss->ad = ss->a + ss->T ; }
  if ( ss->type == 4 || ss->type == 5|| ss->type == 6 ) { ss->ho = ss->ad + ss->T; ss->hd = ss->ho + ss->T ; }
  if ( ss->type == 5 || ss->type == 6 ) { 
    ss->mu = dvector ( 1 , ss->k ) ; 
    ss->alpha = dvector ( 1 , ss->k ) ; 
    ss->logc = ss->hd + ss->T ;  
    ss->logco = ss->hd + ss->T ;  
    ss->logcd = ss->hd + 2*ss->T ;  
  }
  ss->w = dvector ( 1 , ss->G ) ; 
  ss->time = dvector ( 1 , ss->G ) ; 
  ss->lz = dvector ( 1 , ss->G ) ; 
  ss->s = imatrix ( 1 , ss->G , 1 , 5 ) ; /* array of scores */
  if ( ss->find_ebs ) {
    allocate_dmatrix_family ( &(ss->A) , ss->k , 1 ) ; 
  }
  if ( ss->make_eb_table ) {
    ss->meetings = imatrix ( 1 , ss->T , 1 ,  ss->T ) ;
  }
  ss->pg = dmatrix ( 0 , ss->maxg , 0 , ss->maxg ) ; 
  ss->tpg = dmatrix ( 0 , ss->maxg , 0 , ss->maxg ) ; 
  if ( ss->score_table ) {
    ss->og = dmatrix ( 0 , ss->maxg , 0 , ss->maxg ) ; 
    ss->lg = dmatrix ( 0 , ss->maxg , 0 , ss->maxg ) ; 
  }

  return status ; 
}

static int free_space ( opt_args *s  ) {
  int status = 0 ; 

 free_dvector (  s->a , 1 , s->k ) ; 
  free_dvector ( s->w , 1 , s->G ) ; 
  free_dvector ( s->time , 1 , s->G ) ; 
  free_dvector ( s->lz , 1 , s->G ) ; 
  free_imatrix ( s->s,  1 , s->G , 1 , 4 ) ; /* array of scores */
  if ( s->find_ebs ) { 
    free_dmatrix_family ( &(s->A) ) ; 
  }

  return status ; 
}

static int read_tms ( opt_args *ss  ) {
  FILE    *fp;
  char *file = ss->tfile ;
  int t , t1 , status = 0 ; 
  fp = fopen( file, "r" );
  if( !fp )   fprintf( stderr, "No such file: %s\n", file ), exit(0);
  printf ( "Reading teams from %s... ", file ) ; fflush ( stdout ) ; 
  for ( t = 1 ; t <= ss->T ; t ++ ) {
    if ( fscanf(fp,"%d %s" , &t1 , ss->tms[t] ) == EOF ) {
      status = -1 ;       break ;
    }

  }
  if ( status < 0 ) { 
    fprintf ( stderr , "error reading line %d of file %s\n" , 
			       t , file ) ; 
  }
  printf ( "done\n" ) ;

  return status ; 
}


static int read_scores ( opt_args *ss  ) {
  /* reads in each line of the file */
  /* one error-handling strategy would be:
     if any team appears twice on one row
     or is outside 1,T then generate warning, 
     ditch that row, and reduce number of games 

     the current version crashes on error */

  int status = 0 ; 
  FILE    *fp;
  char *file = ss->sfile ;
  int g , t1 , s1 , t2 , s2; 
  double *w = ss->w ; 
  double *time = ss->time ; 
  int **s = ss->s , tots = 0 ; 
  
  fp = fopen( file, "r" );
  if( !fp )   fprintf( stderr, "No such file: %s\n", file ), exit(0);
  printf ( "Reading scores from %s... ", file ) ; fflush ( stdout ) ; 

  for ( g = 1 ; g <= ss->G ; g ++ ) {
    if ( fscanf(fp,"%d %d %d %d " , &t1 , &s1 , &t2 , &s2 ) == EOF ) {
      status = -1 ;       break ;
    }
    if ( ss->times_read ) {
      if ( fscanf(fp,"%lf " , &time[g] ) == EOF ) {
	status = -1 ;       break ;}
      if ( time[g] <= 0.0 ) { fprintf ( stderr , "bizarre time %g\n", time[g] ) ; status -- ; }
    } else time[g] = 1.0 ; 
    if ( ss->weighted ) {
      if ( fscanf(fp,"%lf " , &w[g] ) == EOF ) {
	status = -1 ;       break ;}
      if ( w[g] < 0.0 ) { fprintf ( stderr , "bizarre weight %g\n", w[g] ) ; status -- ; }
    }
    s[g][1] = t1 ; s[g][2] = s1 ; s[g][3] = t2 ; s[g][4] = s2 ; s[g][5] = s1 + s2 ; 
    tots += s1 + s2 ; 
    if ( ( t1 > ss->T ) ||  ( t1 < 1 )  ||  ( t1 > ss->T ) ||  ( t1 < 1 ) ) {
      fprintf ( stderr , "team number out of range 1,%d\n", ss->T ) ; 
      status -- ; 
    }
    if ( ( t1 == t2 ) ) {
      fprintf ( stderr , "team playing itself? %d\n", t1 ) ; 
      status -- ; 
    }
    if ( s1 < 0 || s2 < 0 ) {
      fprintf ( stderr , "bizarre scores %d %d\n", s1 , s2 ) ; 
      status -- ; 
    }
  }
  if ( status < 0 ) { 
    fprintf ( stderr , "error reading line %d of file %s\n" , 
			       g , file ) ; 
  }
  printf ( "done\n" ) ;
  ss->total_scores = tots ; 

  return status ; 
}

#define STYLE %5.3f
static void print_state ( opt_args *ss , FILE *fp ) { /* the main output routine */
  int i , T = ss->T ; 
  double d1 , d2 , d3 , d4 ; 
  for ( i = 1 ; i <= ss->T ; i++ ) {
    fprintf( fp , "%3d " , i ) ;
    if ( ss->readt ) fprintf( fp , "%10s " , ss->tms[i] ) ;
    if ( ss->type == 2 || ss->type == 4 || ss->type == 5  || ss->type == 6 ) {
      fprintf( fp , "%5.3f %5.3f" , 
	      (d1 = ss->reportfact * ss->ao[i] + ss->reportmean) ,
	      (d2 = ss->reportfact * ss->ad[i] + ss->reportmean) ) ;
      if ( ss->type == 4 || ss->type == 5  || ss->type == 6  ) {
	fprintf( fp , " %5.3f %5.3f" , 
		(d3 = ss->reportfact * ss->ho[i] + ss->reportmean) ,
		(d4 = ss->reportfact * ss->hd[i] + ss->reportmean) ) ;
	fprintf( fp , " %6.2f " , 	      d1+d2+d3+d4 ) ;
      }    else {
	fprintf( fp , " %6.2f" , 	      d1+d2 ) ;
      }
    } 
    if ( ss->type == 1 ) {
      fprintf( fp , "%6.3f" , 
	      ss->reportfact * ss->a[i] + ss->reportmean  ) ;
    }
    if ( ss->type == 5  ) {
      fprintf( fp , " %5.3f" , ss->logc[i] ) ; 
    }
    if ( ss->type == 6  ) {
      fprintf( fp , " %5.3f %5.3f" , ss->logco[i] , ss->logcd[i] ) ; 
    }
    if ( ss->find_ebs && ss->A.in_computed ) {
      if ( ss->type == 2 || ss->type == 4 || ss->type == 5  || ss->type == 6 ) {
	fprintf( fp , " %5.3f" , sqrt(ss->A.in[i][i] ) * ss->reportfact  )  ;
	fprintf( fp , " %5.3f" , sqrt(ss->A.in[i+T][i+T] ) * ss->reportfact  )  ;
	if ( ss->type == 4 || ss->type == 5  || ss->type == 6 ) {
	  fprintf( fp , " %5.3f" , sqrt(ss->A.in[i+2*T][i+2*T] ) * ss->reportfact  )  ;
	  fprintf( fp , " %5.3f" , sqrt(ss->A.in[i+3*T][i+3*T] ) * ss->reportfact  )  ;
	  if ( ss->type == 6 ) {
	    fprintf( fp , " %5.3f" , sqrt(ss->A.in[i+4*T][i+4*T] ) * ss->reportfact  )  ;
	    fprintf( fp , " %5.3f" , sqrt(ss->A.in[i+5*T][i+5*T] ) * ss->reportfact  )  ;
	  }
	}
      } else {
	fprintf( fp , " %5.3f" , sqrt(ss->A.in[i][i] ) * ss->reportfact  )  ;
      }
    }
    fprintf( fp , "\n" );
  }
} 
static int poisson ( double l , double *p , int maxg ) { 
  int g = 0 , mpg = 0 ; 
  double pr , maxpr = 0.0 ;
  pr = exp ( - l ) ; 
  p[-1] = 0.0 ; 
  for ( ; g <= maxg ; ) {
    p[-1] += pr ;
    p[g] = pr ; 
/*    printf ( " p[%d] = %9g " , g , pr ) ;  */
    if ( pr > maxpr ) { mpg = g ; maxpr = pr ; }
    g ++ ; 
    pr *= ( l / (double)(g) ) ;
  }
/*  printf ( " sum = %9g \n" , p[-1] ) ;  */
  return mpg ;  /* most probable goal number */
}
static int gamma_poisson ( double c , double logc , double l , double logla , double *p , int maxg , double *logfact ) { 
  int g = 0 , mpg = 0 ; 
  double pr , maxpr = 0.0 ,  lamc , lpr , dg , loglamc , sc ;
  double c_const = - gammln(c) + c * logc ;

  lamc = l + c ; loglamc  = log(lamc) ; 
  p[-1] = 0.0 ; 
  for ( ; g <= maxg ; ) {
    dg = (double)(g) ; 
    sc = dg + c ; 
    lpr = c_const + dg * logla + gammln ( sc ) - logfact[g] - sc * loglamc ;
    pr = exp ( lpr ) ;
    p[-1] += pr ;
    p[g] = pr ; 
/*    printf ( " p[%d] = %9g " , g , pr ) ;  */
    if ( pr > maxpr ) { mpg = g ; maxpr = pr ; }
    g ++ ; 
  }
  if ( fabs( p[-1] - 1.0 ) > 0.01 ) printf ( " sum = %9g \n" , p[-1] ) ;  
  return mpg ;  /* most probable goal number */
}
static double odds ( double p , opt_args *s  ) { 
  if ( p < 1.0 && p > 0.0 ) { 
    return (( 1.0 + s->tax - p ) / p) ; 
  } else {
/*    fprintf ( stderr , "odds: p out of range!\n" ) ; */ return 0.0 ;
  } 
}
static void print_odd_head ( FILE *fp ) {
  fprintf( fp , "\n== Odds: the author disclaims all liabilities for loss or injury\n=  resulting from use of these numbers.\n=  No representation is made as to the fitness of\n=  these numbers for any purpose \n" );
  if ( fp )   fprintf( fp , "\n== If these odds are correct then one would expect to make money\n=  when a bookie offers bigger odds for some event. Mean expected\n=  winnings (a possible way of choosing between favourable bets)\n=  are p ( O+1 ) - 1  where O=THEIR ODDS.\n=  This equals (O+1)/(o+1) - 1 , where o = our odds\n" );
  if ( fp )   fprintf( fp , "\n== If taxes are included when betting then \n=  Mean expected winnings are [ p ( O+1 ) - (1+t) ] / (1+t) per unit bet     \n=  If winnings taxed then     [ p ( O+1 ) / (1+t) - 1 ]   where O=THEIR ODDS.\n=  So favourable odds are given by (O+1) >= (1+t)/p -> O >= (1+t)/p - 1      \n=  = ( 1 + t - p ) / p . This is the quantity reported as \"our odds\"\n\n" );
}
static void print_odds ( opt_args *s , FILE *fp , FILE *fpl ) 
{
     /* human = human-readable */
  int junk , x1 , x2 ,t1,t2 , T = s->T , mpg1 , mpg2 , k , maxg = s->maxg ; 
  double *ao = s->a , *ad = s->a + s->T ; 
  double *ho = s->a + 2*s->T , *hd = s->a + 3*s->T ; 
  double *logc = s->a + 4*s->T ; 
  double *logco = s->a + 4*s->T ; 
  double *logcd = s->a + 5*s->T ; 
  double time , lt , lambda1 , lambda2 , ll1 , ll2 , pmps ; 
  double *scratch , *p1 , *p2 , win , lose , draw , sdraw ; 
  double *v1, *v2 , var , sdl1 , sdl2 , factor ; 
  double   llc ,  lc , lc1 , lc2 , c , c1 , c2 , thislambda1 , thislambda2 , step , sdc1 , sdc2 ; 
int type = s->type ; 
  if ( s->type < 4 ) { ho = ao ; hd = ad ; }
  p1 = dvector ( -1 , maxg ) ; /* poisson dist */
  p2 = dvector ( -1 , maxg ) ; 
  scratch = dvector ( -1 , maxg ) ; 
  v1 = dvector ( 1 , s->k ) ; 
  v2 = dvector ( 1 , s->k ) ;  /* vectors for finding error bars */
  if ( s->verbose >= 1 ) { printf ("computing odds... " ) ; fflush (stdout);}
  if ( fp ) print_odd_head ( fp ) ;
  for ( t1 = 1 ; t1 <= T ; t1 ++ ) {
    if ( s->readt && fp )
      fprintf( fp , "\n== %s at home ========================\n\n" , s->tms[t1] );
    for ( t2 = 1 ; t2 <= T ; t2++ ) {
      if ( t1 != t2 ) {
	time = 1.0 ; /* note standard duration assumed */
	lt = log(time) ;
	if ( s->type == 2 || s->type == 4 || s->type == 5|| s->type == 6 ) {
	  ll1 = lt + ( ho[t1] - ad[t2] ) ; 
	  ll2 = lt + ( ao[t2] - hd[t1] ) ; 
	  lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
	  lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */
	  if ( s->type == 5 ) { /* type 5! */
	    lc1 = logc[t1] ;     lc2 = logc[t2] ; lc = lc1+lc2 ;  c = exp(lc) ;
	       llc = lambda1 + lambda2 + c ; 
	  }
	  if ( s->type == 6 ) { /* type 6! */
	    lc1 = logco[t1] + logcd[t2] ; /* consistency of score 1 */
	    lc2 = logco[t2] + logcd[t1] ;
	    c1 = exp(lc1) ;	    c2 = exp(lc2) ;
	  }
	} else /* if ( s->type == 1 ) */ {
	  fprintf ( stderr , "invalid type for odds, sorry\n" ) ; 
	}

	if ( s->type == 2 || s->type == 4 ) {
	  mpg1 = poisson ( lambda1 , p1 , maxg ) ; 
	  mpg2 = poisson ( lambda2 , p2 , maxg ) ; 
	  pmps = p1[mpg1] * p2[mpg2] ; 
	} 	  else if ( s->type == 6 )  {
	  mpg1 = gamma_poisson ( c1 , lc1 , lambda1 , ll1 , p1 , maxg , s->logfact  ) ; 
	  mpg2 = gamma_poisson ( c2 , lc2 , lambda2 , ll2 , p2 , maxg , s->logfact ) ; 
	  pmps = p1[mpg1] * p2[mpg2] ;
	} /* no equivalent operation for type 5 */

	if ( s->find_ebs && s->A.in_computed ) {
	  for ( k = 1 ; k <= s->k ; k ++ ) {
	    v1[k] = v2[k] = 0.0 ;
	  } 
	  v1[t1+2*T] = v2[t1+2*T] = 1.0 ; /* home o */
	  v1[t2+T] = v2[t2+T] = -1.0 ;    /* away d */
	  var =  lu_quadratic_form ( v1, &(s->A) , v2 ) ;
	  if ( var > 0.0 ) sdl1 = sqrt(var) ; else { fprintf ( stderr , "warning, negative variance %g\n" , var ) ; }
	  for ( k = 1 ; k <= s->k ; k ++ ) {
	    v1[k] = v2[k] = 0.0 ;
	  } 
	  v1[t1+3*T] = v2[t1+3*T] = 1.0 ; /* home d */
	  v1[t2] = v2[t2] = -1.0 ;    /* away o */
	  var =  lu_quadratic_form ( v1, &(s->A) , v2 ) ; 
	  if ( var > 0.0 ) sdl2 = sqrt(var) ; else { fprintf ( stderr , "warning, negative variance! %g\n" , var ) ; }
	  if ( type == 6 ) { /* get ebs for c1 and c2 too */
	    for ( k = 1 ; k <= s->k ; k ++ ) {
	      v1[k] = v2[k] = 0.0 ;
	    } 
	    v1[t1+4*T] = v2[t1+4*T] = 1.0 ; /* home o */
	    v1[t2+5*T] = v2[t2+5*T] = 1.0 ; /* away d */
	    var =  lu_quadratic_form ( v1, &(s->A) , v2 ) ;
	    if ( var > 0.0 ) sdc1 = sqrt(var) ; else { fprintf ( stderr , "warning, negative variance %g\n" , var ) ; sdc1 = 0.0 ; }
	    for ( k = 1 ; k <= s->k ; k ++ ) {
	      v1[k] = v2[k] = 0.0 ;
	    } 
	    v1[t1+5*T] = v2[t1+5*T] = 1.0 ; /* home d */
	    v1[t2+4*T] = v2[t2+4*T] = 1.0 ; /* away o */
	    var =  lu_quadratic_form ( v1, &(s->A) , v2 ) ; 
	    if ( var > 0.0 ) sdc2 = sqrt(var) ; else { fprintf ( stderr , "warning, negative variance! %g\n" , var ) ; sdc2 = 0.0 ; }
	  } 
	}

	win = 0.0 ; draw = 0.0 ; lose = 0.0 ;  sdraw = 0.0 ; 
	constantdmatrix ( s->pg , 0,maxg,0,maxg,0.0);
	if ( s->MAP ) { /* compute all the win lose draw stats using MAP */
	  if ( s->type == 2 || s->type == 4  || s->type == 6 ) 
	    increment_poisson_product ( s->pg ,  p1 , p2 , &win, 
				       &lose, &draw, &sdraw, maxg , 
				       1.0 , &pmps , &mpg1 , &mpg2 ) ;
	  else if ( s->type == 5 ) 
	    increment_gamma_product ( s , s->pg , llc , lc , c , ll1 , ll2 , &win, &lose, &draw, 
				     &sdraw, maxg , 1.0 , &pmps , 
				     &mpg1 , &mpg2 ) ;
	}
	else 	if ( s->find_ebs && s->A.in_computed ) { /* types 2,4,6 only */
  /* not MAP:  do a rough sample over the error bars */
	  if ( ( s->eb_method == 1 ) || ( type != 6 ) ) {
	    factor = 0.2 ;
	    step = sqrt(5.0/4.0) ; 
	    increment_poisson_product ( s->pg ,  p1 , p2 , &win, &lose, &draw, &sdraw, maxg , factor  , &pmps , &mpg1 , &mpg2  ) ;
	    for ( x1 = -1 ; x1 <= 1 ; x1 += 2 ) {
	      thislambda1 =  lambda1 * exp ((double)(x1) * sdl1 * step ) ; 
	      junk =  ( type == 6 ) ? 
		gamma_poisson ( c1 , lc1 /* log (c1) */ , thislambda1 , log(thislambda1) , p1 , maxg , s->logfact  ) :
		  poisson ( thislambda1 , p1 , maxg ) ; 
	      for ( x2 = -1 ; x2 <= 1 ; x2 += 2 ) {
		thislambda2 = lambda2 * exp ((double)(x2) * sdl2 * step ) ;
		junk = (type == 6) ?
		  gamma_poisson ( c2 , lc2 , thislambda2 , log(thislambda2) , p2 , maxg , s->logfact ) :
		    poisson ( thislambda2  , p2 , maxg ) ; 
		increment_poisson_product ( s->pg ,  p1 , p2 , &win, &lose, &draw, &sdraw, maxg , factor  , &pmps , &mpg1 , &mpg2  ) ;
	      }
	    }
	  } else { /* this for type 6 */ /* eb_method == 2 */
	    /* tot up the marginals over goals and then do a single increment */
	    marginal_p ( p1 , lambda1 , sdl1 , c1 , sdc1 , scratch , maxg , s->logfact ) ; 
	    marginal_p ( p2 , lambda2 , sdl2 , c2 , sdc2 , scratch , maxg , s->logfact ) ; 
	    increment_poisson_product ( s->pg ,  p1 , p2 , &win, &lose, &draw, &sdraw, maxg , 1.0 , &pmps , &mpg1 , &mpg2  ) ;
	  }
	} else {
	  fprintf ( stderr , "Eek , can't do requested non-MAP without error bars!\n" ) ; 
	}

	if ( fp ) {
	  if ( s->readt ) {
	    fprintf( fp , "%s v " , s->tms[t1] ) ;
	    fprintf( fp , "%-5s" , s->tms[t2] ) ;
	  } 
	  fprintf( fp , " (%2d v %-2d) " , t1 , t2 ) ;
	} 
	if ( fpl ) {
	  if ( s->readt ) {
	    fprintf( fpl , "%-5s " , s->tms[t1] ) ;
	    fprintf( fpl , "%-5s " , s->tms[t2] ) ;
	  } 
	  fprintf( fpl , "%-2d %-2d " , t1 , t2 ) ;
	}
	if ( fp ) 
	  fprintf( fp , "mp score %d : %d " , mpg1 , mpg2 ) ;
	if ( fp && s->mpscore_odds ) {
	  fprintf ( fp , " Pr %-6.3g Odds %-6.3g" , 
		  pmps , odds( pmps ,s )  ) ;
	  if ( s->odd_logp ) 
	    fprintf ( fp , " logP %-6.3g" , 
		   log(pmps) ) ;
	  fprintf ( fp , "\n" ) ; 
	}

	if ( fp && s->score_draw ) {
	  fprintf ( fp , " Score draw Pr %-6.3g  Odds %-6.3g" , 
 		   sdraw , odds( sdraw ,s )  ) ;
	  if ( s->odd_logp ) 
	    fprintf ( fp , " logP %-6.3g" ,  log(sdraw) ) ;
	  fprintf ( fp , "\n" ) ; 
	}

	if ( fpl )  
	  fprintf( fpl , "%d %d " , 
		  mpg1 , mpg2 ) ;
	if ( fpl && s->mpscore_odds )  
	  fprintf( fpl , "%-6.3g %-6.3g %-6.3g " , 
		   pmps , odds( pmps,s ) , log(pmps) ) ;
	if ( fpl && s->score_draw ) 
	  fprintf ( fpl , "%-6.3g %-6.3g %-6.3g " ,
		  sdraw , odds( sdraw,s ) , log( sdraw ) ) ;
	if ( fp ) {
	  if ( s->MAP ) { fprintf( fp , "      [MAP" ) ; }
	  else { fprintf( fp , "      [Marg" ) ; }
	  fprintf ( fp , "%d] " , s->type ) ; 
	  fprintf( fp , "win lose draw %4.2f %4.2f %4.2f   Odds %5.1f %5.1f %5.1f" , 
		   win , lose , draw , odds(win,s)  , odds(lose,s) , odds(draw,s)  ) ;
	  if ( s->odd_logp ) 
	    fprintf ( fp , " - LogP %5.2g %5.2g %5.2g" ,  log(win), log(lose) , log(draw) ) ;
	  fprintf ( fp , "\n" ) ; 
	}
	if ( fpl ) 
	   fprintf( fpl , "%4.2f %4.2f %4.2f %5.1f %5.1f %5.1f %5.2g %5.2g %5.2g " , 
		  win , lose , draw , odds(win,s)  , odds(lose,s) , odds(draw,s)  , log(win), log(lose) , log(draw)  ) ;
/* error bars */
	if ( s->find_ebs && s->A.in_computed ) {
	  if ( fp ) 
	    fprintf( fp , "      log lambda1 = %5.1f +/- " , ll1 ) ; 
	  if ( fpl ) 
	    fprintf( fpl , "%5.1f " , ll1 ) ; 
	  if ( fp )  fprintf ( fp , "%4.2g " ,  sdl1 ) ; 
	  if ( fpl ) fprintf ( fpl , "%4.2g " ,  sdl1 ) ; 
	  
	  if (fp) fprintf( fp , "  log lambda2 = %5.1f +/- " , ll2 ) ; 
	  if ( fpl ) fprintf( fpl , "%5.1f " , ll2 ) ; 
	  if ( fp ) fprintf ( fp , "%4.2g" ,  sdl2 ) ; 
	  if ( fpl )	  fprintf ( fpl , "%4.2g " ,  sdl2 ) ; 
	  
	  if ( fp ) fprintf ( fp , "\n" ) ; 
	}
	if ( fp ) fprintf ( fp , "\n" ) ; 

	/* print out a wee table using MAP parameters */
	if ( s->score_table ) {
	  pg_process ( s ) ; 
	  if ( fp ) pretty_print_score_table ( s , fp ) ; 
	  if ( fpl ) ugly_print_score_table ( s , fpl ) ; 
	  if ( fp ) fprintf ( fp , "\n" ) ; 
	}

/* end of the t1 - t2 encounter */
	if ( fpl ) fprintf ( fpl , "\n" ) ; 
      }
    }
  }
  if ( s->verbose >= 1 ) { printf ("done\n" ) ; fflush (stdout);}
} 

void marginal_p ( double *p , double lam , double sdl , double c , double sdc , double *scratch , int maxg , double *logfact ) 
{
  int g , x , y ; 
  double thisl , thisc , factor = 0.2 , step = sqrt(5.0/4.0) ; 
  gamma_poisson ( c , log (c) , lam , log(lam) , scratch , maxg , logfact ) ;
  for ( g = 0 ; g <= maxg ; g ++ ) {
    p[g] = scratch[g] * factor ; 
  }
  for ( x = -1 ; x <= 1 ; x += 2 ) {
    thisl = lam * exp ((double)(x) * sdl * step ) ;
    for ( y = -1 ; y <= 1 ; y += 2 ) {
      thisc = c * exp ((double)(y) * sdc * step ) ;
      gamma_poisson ( thisc , log (thisc) , thisl , log(thisl) , scratch , maxg , logfact ) ;
      for ( g = 0 ; g <= maxg ; g ++ ) {
	p[g] += scratch[g] * factor ; 
      }
    }
  }
}

static void  ugly_print_score_table  ( opt_args *s , FILE *fp  ) {
  int g1,g2 ;
  double **pg = s->pg ; 
  double **og = s->og ; 
  double **lg = s->lg ;
  /* tables */
  for ( g1 = 0 ; g1 <= s->maxgr ; g1 ++ ) 
    for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
      fprintf ( fp , "%4.2f " , pg[g1][g2] ) ; 
  for ( g1 = 0 ; g1 <= s->maxgr ; g1 ++ ) 
    for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
      odds_print ( fp , og[g1][g2] ) ; 
  for ( g1 = 0 ; g1 <= s->maxgr ; g1 ++ ) 
    for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
      fprintf ( fp , "%5.3g " , lg[g1][g2] ) ; 
} 

static void  pretty_print_score_table  ( opt_args *s , FILE *fp  ) {
  int g1,g2 ;
  double **pg = s->pg ; 
  double **og = s->og ; 
  double **lg = s->lg ; 

  /* headers */
  if ( !( s->odd_logp) ) 
    fprintf ( fp , "      " ) ; /* space at start */
  fprintf ( fp , "Score Probabilities, Odds" ) ; 
  if ( s->odd_logp ) fprintf ( fp , ", and Log Probs" ) ; 
  fprintf ( fp , " (home score = y axis; away = x axis)\n" ) ; 
  if ( !( s->odd_logp) ) 
    fprintf ( fp , "      " ) ; /* space at start */
  fprintf ( fp , "   " ) ; 
  for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
    fprintf ( fp , "%4d " , g2 ) ; 
  if ( !( s->odd_logp) ) {
    fprintf ( fp , "  " ) ; /* space */
    fprintf ( fp , "   " ) ;  /* home goal again */
  }
  for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
    fprintf ( fp , "%5d " , g2 ) ; 
  if ( s->odd_logp ) {
    for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
      fprintf ( fp , "%5d " , g2 ) ; 
  }
  fprintf ( fp , "\n" ) ; 

  /* tables */
  for ( g1 = 0 ; g1 <= s->maxgr ; g1 ++ ) {
    if ( !( s->odd_logp) ) 
      fprintf ( fp , "      " ) ; /* space at start */
    fprintf ( fp , "%-2d " , g1 ) ;
    for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
      fprintf ( fp , "%4.2f " , pg[g1][g2] ) ; 
    if ( !( s->odd_logp) ) {
      fprintf ( fp , "  " ) ; /* space */
      fprintf ( fp , "%-2d " , g1 ) ; /* home goal again */
    }
    for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
      odds_print ( fp , og[g1][g2] ) ; 
    if ( s->odd_logp ) {
      for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) 
	fprintf ( fp , "%5.3g " , lg[g1][g2] ) ; 
    }
    fprintf ( fp , "\n" ) ; 
  }
} 
static void odds_print ( FILE * fp , double o ) {
  if ( o > 999.9 ) 
    fprintf ( fp , "%5.1f " , 999.9 ) ; 
  else 
    fprintf ( fp , "%5.1f " , o ) ; 
}
static void  pg_process ( opt_args *s ) {
  int g1,g2 ;
  double **pg = s->pg ; 
  double **og = s->og ; 
  double **lg = s->lg ; 

  for ( g1 = 0 ; g1 <= s->maxgr ; g1 ++ ) {
    for ( g2 = 0 ; g2 <= s->maxgr ; g2 ++ ) {
      og[g1][g2] = odds (  pg[g1][g2] ,s ) ; 
      lg[g1][g2] = log (  pg[g1][g2] ) ; 
    }
  }
} 
static void increment_poisson_product ( double **pg , double *p1 , 
 double *p2 , double *win , double *lose , 
		 double *draw, double *sdraw , int maxg , double weight,
	       double *pmps , int *mpg1 , int *mpg2  ) { 
  double tot = 0.0 , this  ;
  int g1,g2 ;

  *pmps = 0.0 ; 
  for ( g1 = 0 ; g1 <= maxg ; g1 ++ ) {
    for ( g2 = 0 ; g2 <= maxg ; g2 ++ ) {
      this = p1[g1] * p2[g2] * weight ;
      tot += this ;
      if ( g1 > g2 ) *win += this ; 
      else if ( g1 < g2 ) *lose += this ;
      else {
	*draw += this ; 
	if ( g1 > 0 ) 
	  *sdraw += this ;
      }

      pg[g1][g2] += this ; 
      if ( *pmps < pg[g1][g2] ) { *pmps = pg[g1][g2]  ; *mpg1 = g1 ; *mpg2 = g2 ; }
    }
  }
}
static void increment_gamma_product ( opt_args *s , double **pg , double llc , double lc ,
				     double c , double ll1 , double ll2 , 
 double *win , double *lose , 
		 double *draw, double *sdraw , int maxg , double weight,
	       double *pmps , int *mpg1 , int *mpg2  ) { 
  double tot = 0.0 , this  ;
  int g1,g2 ; double s1 , s2 , ssc ; 
  double logllc = log(llc) ; 
  double c_const = - gammln(c) + c * lc ;
  double *logfact = s->logfact ; /* factorial */
  double **tpg = s->tpg ; 

  *pmps = 0.0 ; 
  for ( g1 = 0 ; g1 <= maxg ; g1 ++ ) {
    for ( g2 = 0 ; g2 <= maxg ; g2 ++ ) {
      s1 = (double) g1 ;       s2 = (double) g2 ; 
      ssc = s1 + s2 + c ; 
      this  = s1 * ll1  ;    this += s2 * ll2  ; 
      this -= logfact[g1] + logfact[g2] ; 
      this += gammln (ssc) + c_const - ssc * logllc ; 
      this = exp(this) ; 
      tpg[g1][g2] = this ; 
      tot += this ;
    }
  }
  if ( fabs(tot-1.0) > 0.01 ) { printf ( "tot = %g\n" , tot ) ; }
  tot = 1.0 / tot ; 
  for ( g1 = 0 ; g1 <= maxg ; g1 ++ ) {
    for ( g2 = 0 ; g2 <= maxg ; g2 ++ ) {
      tpg[g1][g2] *= tot ; 
      this = tpg[g1][g2] ;
      if ( g1 > g2 ) *win += this ; 
      else if ( g1 < g2 ) *lose += this ;
      else {
	*draw += this ; 
	if ( g1 > 0 ) 
	  *sdraw += this ;
      }

      pg[g1][g2] += this ; 
      if ( *pmps < pg[g1][g2] ) { *pmps = pg[g1][g2]  ; *mpg1 = g1 ; *mpg2 = g2 ; }
    }
  }
}

static void wld ( double *p1 , double *p2 , double *win , double *lose , 
		 double *draw , double *sdraw , int maxg ) {
  double tot = 0.0 , this  ;
  int g1,g2 ;
  *win = 0.0 ; *draw = 0.0 ; *lose = 0.0 ; *sdraw = 0.0 ; 
  for ( g1 = 0 ; g1 <= maxg ; g1 ++ ) {
    for ( g2 = 0 ; g2 <= maxg ; g2 ++ ) {
      this = p1[g1] * p2[g2] ;
      tot += this ;
      if ( g1 > g2 ) *win += this ; 
      else if ( g1 < g2 ) *lose += this ;
      else { 
	*draw += this ; 
	if ( g1 > 0 ) *sdraw += this ; 
      }
    }
  }
}

static void diagnose ( opt_args *ss , FILE *fp ) { /* checks each game to see 
						      how big a discrepancy */
  int  g,t1,t2 ;
  double /* f=0.0 ,*/ w, a1, a2 ,s1,s2 , stot, p1 , p2 , gr1 , gr2 ;
  int **s = ss->s ; 
  double *a = ss->a  ; 
  double *lz = ss->lz ; 

  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; a1=a[t1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; a2=a[t2] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      s1 *= w ; 
      s2 *= w ; 
    }
    lz[g] = logz ( a1 , a2 ) ; 
    stot = s1+s2 ; 
    if ( a1 < a2 ) { p1 = exp ( a1 -  lz[g] ) ; p2 = 1.0 - p1 ; }
    else { p2 = exp ( a2 -  lz[g] ) ; p1 = 1.0 - p2 ; }
/*       f += s1 * a1 + s2 * a2 - ( s1 + s2 ) * lz[g] ; */
/* so    df/da1 += s1 - (s1+s2) * p1                    */
    gr1 = s1 - stot * p1 ; /* minus because f -> -f */
    gr2 = s2 - stot * p2 ; 

    fprintf( fp , "%3d: %8.2g %8.2g " , g , gr1 , gr2 ) ;
    if ( ss->readt ) fprintf( fp , "%10s " , ss->tms[t1] ) ;
    else  fprintf( fp , "%2d " , t1 ) ;
    fprintf( fp , "%2.0f %5.2f  " , s1 ,  stot * p1 ) ;
    if ( ss->readt ) fprintf( fp , "%10s " , ss->tms[t2] ) ;
    else  fprintf( fp , "%2d " , t2 ) ;
    fprintf( fp , "%2.0f %6.1f\n" , s2 ,  stot * p2 ) ;

    /* things to scatter plot */
  }
} 

double logprob(double *x,   void *arg )
{
  int i , g,t1,t2 ;
  double f=0.0 , w, a1, a2 ,s1,s2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *a = x ; 
  double *lz = ss->lz ; 
 
  if ( ss->verbose >= 2 ) 
    {  printf ( "log prob. at " ) ;
       for ( i = 1 ; i <= ss->k ; i++ ) printf( "%g " , x[i] ) ;
     } 
  ss->logprior = 0.0 ; 
  for ( i = 1 ; i <= ss->k ; i++ ) ss->logprior -= x[i]*x[i]* ss->priorfactor ;
  /* run through games */
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; a1=a[t1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; a2=a[t2] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      s1 *= w ; 
      s2 *= w ; 
    }
    lz[g] = logz ( a1 , a2 ) ; 
    f += s1 * a1 + s2 * a2 - ( s1 + s2 ) * lz[g] ;
  }
  ss->loglike = f ; 
  f += ss->logprior ; 
  if ( ss->verbose >= 2 ) 
    printf(" : %g\n",f);
  else if ( ss->verbose >= 1 ) 
    printf("logprob1 : %g  loglike : %g\n",f,  ss->loglike );
  return(-f); 
}

double logz ( double a1 , double a2 ) {
  double lz , ama , ami ; 
  if ( a1 > a2 ) { ama = a1 ; ami = a2 ; }
  else { ama = a2 ; ami = a1 ; }
/*  lz = log (  exp(ama) + exp(ami) ) 
       = ama + log ( 1 + exp(ami-ama) ) */
  lz = ama + log ( 1.0 + exp ( ami - ama ) ) ;
  return lz ; 
}

void grad_logprob(double *x, double *gr, void *arg )
{
  int i , g,t1,t2 ;
  double /* f=0.0 ,*/ w, a1, a2 ,s1,s2 , stot, p1 , p2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *a = x  ; 
  double *lz = ss->lz ; 

  if ( ss->verbose >= 2 ) {
    printf ( "gradient at " ) ;
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , x[i] ) ;
  }
  for ( i = 1 ; i <= ss->k ; i++ ) gr[i] = 2.0 * x[i] * ss->priorfactor ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; a1=a[t1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; a2=a[t2] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      s1 *= w ; 
      s2 *= w ; 
    }
    lz[g] = logz ( a1 , a2 ) ; 
    stot = s1+s2 ; 
    if ( a1 < a2 ) { p1 = exp ( a1 -  lz[g] ) ; p2 = 1.0 - p1 ; }
    else { p2 = exp ( a2 -  lz[g] ) ; p1 = 1.0 - p2 ; }
/*       f += s1 * a1 + s2 * a2 - ( s1 + s2 ) * lz[g] ; */
/* so    df/da1 += s1 - (s1+s2) * p1                    */
    gr[t1] -= s1 - stot * p1 ; /* minus because f -> -f */
    gr[t2] -= s2 - stot * p2 ; 
  }
  if ( ss->verbose >= 2 ) {
    for ( i = 1 ; i <= ss->T ; i++ ) printf ( "%g " , gr[i] ) ;
    printf ( "\n" ) ;
  }
}

void find_A ( opt_args *ss ) 
{
  int i , g,t1,t2 , j ;
  double  w, a1, a2 ,s1,s2 , stot, p1 , p2 , crossterm ;
  int **s = ss->s ; 
  double *a = ss->a  ; 
  double **A = ss->A.m  ; 
  double *lz = ss->lz ; 

  for ( i = 1 ; i <= ss->T ; i++ ) {
    for ( j = 1 ; j <= ss->T ; j++ )
      A[i][j] = ss->curvature_n ;
    A[i][i] += 2.0 * ss->priorfactor ;
  }
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; a1=a[t1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; a2=a[t2] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      s1 *= w ; 
      s2 *= w ; 
    }
    lz[g] = logz ( a1 , a2 ) ; 
    stot = s1+s2 ; 
    if ( a1 < a2 ) { p1 = exp ( a1 -  lz[g] ) ; p2 = 1.0 - p1 ; }
    else { p2 = exp ( a2 -  lz[g] ) ; p1 = 1.0 - p2 ; }
    crossterm =  (- p1 * p2 * stot ) ; 
    A[t1][t2] += crossterm ; 
    A[t2][t1] += crossterm ; 
    A[t1][t1] += -crossterm ; 
    A[t2][t2] += -crossterm ; 
  }
}

double logprob2(double *x,   void *arg )
{ /* log probability assuming fixed duration game, poisson scoring rates */
  int i , g,t1,t2 ;
  double f=0.0 , w, s1,s2 , lambda1 , lambda2 , time , lt , ll1 , ll2 ;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *ao = x , *ad = x + ss->T ; 
  double thisf ; 
 
  if ( ss->verbose >= 2 ) 
    {  printf ( "log prob. at " ) ;
       for ( i = 1 ; i <= ss->k ; i++ ) printf( "%g " , x[i] ) ;
     } 
  ss->logprior = 0.0 ; 
  for ( i = 1 ; i <= ss->k ; i++ ) ss->logprior -= x[i]*x[i]* ss->priorfactor ;
  /* run through games */
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; s2 = (double) s[g][4] ;
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ao[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - ad[t1] ) ; /* expected scoring rates */
    ll1 = lt + ( ao[t1] - ad[t2] ) ; 
    ll2 = lt + ( ao[t2] - ad[t1] ) ; 
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else w = 1.0 ; 
    thisf  = s1 * ll1 - lambda1 ; 
    thisf += s2 * ll2 - lambda2 ; 
    f += w * thisf ; 
  }
  ss->loglike = f ; 
  f += ss->logprior ; 
  if ( ss->verbose >= 2 ) 
    printf(" : %g\n",f);
  else if ( ss->verbose >= 1 ) 
    printf("logprob2 : %g  loglike : %g\n",f,  ss->loglike );
  return(-f); 
}

void grad_logprob2(double *x, double *gr, void *arg )
{
  int i , g,t1,t2 ;
  double /* f=0.0 ,*/ w, lambda1, lambda2 ,s1,s2 , eps1, eps2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  int T = ss->T ; 
  double *ao = x , *ad = x + T ; 
  double *go = gr , *gd = gr + T ; 
  double time, lt ; 

  if ( ss->verbose >= 2 ) {
    printf ( "gradient at " ) ;
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , x[i] ) ;
  }
  for ( i = 1 ; i <= ss->k ; i++ ) gr[i] = 2.0 * x[i] * ss->priorfactor ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else w = 1.0 ; 
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ao[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - ad[t1] ) ; /* expected scoring rates */
    eps1 = ( s1 - lambda1 ) * w ; 
    eps2 = ( s2 - lambda2 ) * w ; 
    go[t1] -= eps1 ;   /* ao derivative */ /* minus because f -> -f */
    go[t2] -= eps2 ; 
    gd[t2] -= - eps1 ; /* ad derivative */ /* minus because f -> -f */
    gd[t1] -= - eps2 ; 
  }
  if ( ss->verbose >= 2 ) {
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , gr[i] ) ;
    printf ( "\n" ) ;
  }
}

void find_A2 ( opt_args *ss ) 
{
  int i , g,t1,t2 , j ;
  double  w, lambda1, lambda2 ,s1,s2  ;
  int **s = ss->s ; 
  double time ; 
  int T = ss->T ; 
  double *ao = ss->a  ; 
  double *ad = ss->a + T  ; 
  double **A = ss->A.m  ; 

  for ( i = 1 ; i <= ss->T ; i++ ) {
    for ( j = 1 ; j <= ss->T ; j++ ) {
      A[i][j] = ss->curvature_n ;
      A[i+T][j+T] = ss->curvature_n ;
    }
  }
  for ( i = 1 ; i <= ss->k ; i++ ) 
    A[i][i] += 2.0 * ss->priorfactor ;

  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ;  s1 = (double) s[g][2] ;
    t2 = s[g][3] ;  s2 = (double) s[g][4] ;

    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lambda1 = time * exp ( ao[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - ad[t1] ) ; /* expected scoring rates */
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      lambda1 *= w ; lambda2 *= w ; 
    } 

    A[t1][t2+T]   += -lambda1 ; 
    A[t2+T][t1]   += -lambda1 ;     /* o1 - d2 */
    A[t1][t1]     += lambda1 ;      /* ao1-ao1 */
    A[t2+T][t2+T] += lambda1 ;  /* ad2-ad2 */
    
    A[t1+T][t2]   += -lambda2 ; 
    A[t2][t1+T]   += -lambda2 ; 
    A[t1+T][t1+T] += lambda2 ; 
    A[t2][t2]     += lambda2 ; 
  }
}
double logprob4(double *x,   void *arg )
{ /* log probability assuming fixed duration game, poisson scoring rates */
  int i , g,t1,t2 ;
  double f=0.0 , w, s1,s2 , lambda1 , lambda2 , time , lt , ll1 , ll2 ;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *ao = x , *ad = x + ss->T ; 
  double *ho = x + 2*ss->T , *hd = x + 3*ss->T ; 
  double thisf ; 
 
  if ( ss->verbose >= 2 ) 
    {  printf ( "log prob. at " ) ;
       for ( i = 1 ; i <= ss->k ; i++ ) printf( "%g " , x[i] ) ;
     } 
  ss->logprior = 0.0 ; 
  for ( i = 1 ; i <= ss->k ; i++ ) ss->logprior -= x[i]*x[i]* ss->priorfactor ;
  /* run through games */
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ; /* 1 is at HOME */
    t2 = s[g][3] ; s2 = (double) s[g][4] ;
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */
    ll1 = lt + ( ho[t1] - ad[t2] ) ; 
    ll2 = lt + ( ao[t2] - hd[t1] ) ; 
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else w = 1.0 ; 
    thisf  = s1 * ll1 - lambda1 ; 
    thisf += s2 * ll2 - lambda2 ; 
    f += w * thisf ; 
  }
  ss->loglike = f ; 
  f += ss->logprior ; 
  if ( ss->verbose >= 2 ) 
    printf(" : %g\n",f);
  else if ( ss->verbose >= 1 ) 
    printf("logprob4 : %g  loglike : %g\n",f,  ss->loglike );
  return(-f); 
}

void grad_logprob4(double *x, double *gr, void *arg )
{
  int i , g,t1,t2 ;
  double /* f=0.0 ,*/ w, lambda1, lambda2 ,s1,s2 , eps1, eps2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  int T = ss->T ; 
  double *ao = x , *ad = x + T ; 
  double *ho = x + 2*ss->T , *hd = x + 3*ss->T ; 
  double *gao = gr , *gad = gr + T ; 
  double *gho = gr + 2*ss->T , *ghd = gr + 3*ss->T ; 
  double time, lt ; 

  if ( ss->verbose >= 2 ) {
    printf ( "gradient at " ) ;
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , x[i] ) ;
  }
  for ( i = 1 ; i <= ss->k ; i++ ) gr[i] = 2.0 *  x[i] * ss->priorfactor ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else w = 1.0 ; 
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */
    eps1 = ( s1 - lambda1 ) * w ; 
    eps2 = ( s2 - lambda2 ) * w ; 
    gho[t1] -= eps1 ;   /* ao derivative */ /* minus because f -> -f */
    gao[t2] -= eps2 ; 
    gad[t2] -= - eps1 ; /* ad derivative */ /* minus because f -> -f */
    ghd[t1] -= - eps2 ; 
  }
  if ( ss->verbose >= 2 ) {
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , gr[i] ) ;
    printf ( "\n" ) ;
  }
}

void find_A4 ( opt_args *ss ) 
{
  int i , g,t1,t2 , j ;
  double  w, lambda1, lambda2 ,s1,s2  ;
  int **s = ss->s ; 
  double time ; 
  int T = ss->T ; 
  double *ao = ss->a  ;   double *ad = ss->a + T  ; 
  double *ho = ss->a + 2*T , *hd = ss->a + 3*T ; 
  double **A = ss->A.m  ; 

  for ( i = 1 ; i <= ss->T ; i++ ) {
    for ( j = 1 ; j <= ss->T ; j++ ) {
      A[i][j] = ss->curvature_n ;
      A[i+T][j+T] = ss->curvature_n ;
      A[i+2*T][j+2*T] = ss->curvature_n ;
      A[i+3*T][j+3*T] = ss->curvature_n ;
    }
  }
  for ( i = 1 ; i <= ss->k ; i++ ) 
    A[i][i] += 2.0 * ss->priorfactor ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ;  s1 = (double) s[g][2] ;
    t2 = s[g][3] ;  s2 = (double) s[g][4] ;

    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      lambda1 *= w ; lambda2 *= w ; 
    } 

    A[t1+2*T][t2+T]   += -lambda1 ; /* home o 1   away d 2 */
    A[t2+T][t1+2*T]   += -lambda1 ;     /* o1 - d2 */
    A[t1+2*T][t1+2*T]     += lambda1 ;      /* ao1-ao1 */
    A[t2+T][t2+T] += lambda1 ;  /* ad2-ad2 */
    
    A[t1+3*T][t2]   += -lambda2 ; /* home  d 1   away  o 2 */
    A[t2][t1+3*T]   += -lambda2 ; 
    A[t1+3*T][t1+3*T] += lambda2 ; 
    A[t2][t2]     += lambda2 ; 
  }
}

double logprob5(double *x,   void *arg )
{ /* log probability assuming fixed duration game, poisson scoring rates */
  int i , g,t1,t2 ;
  double f=0.0 , w, s1,s2 , lambda1 , lambda2 , time , lt , ll1 , ll2  ;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *ao = x , *ad = x + ss->T ; 
  double *ho = x + 2*ss->T , *hd = x + 3*ss->T ; 
  double *logc = x + 4*ss->T ; 
  double thisf , ssc , llc , stot, lc , lc1 , lc2 , c ; 
  double *mu = ss->mu ; 
  double *alpha = ss->alpha ; 
 
  if ( ss->verbose >= 2 ) 
    {  printf ( "log prob. at " ) ;
       for ( i = 1 ; i <= ss->k ; i++ ) printf( "%g " , x[i] ) ;
     } 
  ss->logprior = 0.0 ; 
  for ( i = 1 ; i <= ss->k ; i++ )
    ss->logprior -= ((x[i]-mu[i])*(x[i]-mu[i]) ) * alpha[i] * 0.5 ;
  /* run through games */
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ; /* 1 is at HOME */
    t2 = s[g][3] ; s2 = (double) s[g][4] ; stot = s1 + s2 ; 
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */
    ll1 = lt + ( ho[t1] - ad[t2] ) ; 
    ll2 = lt + ( ao[t2] - hd[t1] ) ; 
/* type 5! */
    lc1 = logc[t1] ;     lc2 = logc[t2] ; lc = lc1+lc2 ;  c = exp(lc) ;
    ssc = stot + c ;    llc = lambda1 + lambda2 + c ; 
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else  w = 1.0 ; /* I have not checked out weightings for type 5 */
    thisf  = s1 * ll1  ; 
    thisf += s2 * ll2  ; 
    thisf += gammln (ssc) - gammln(c) + c * lc - ssc * log(llc) ; 
    f += w * thisf ; 
  }
  ss->loglike = f ; 
  f += ss->logprior ; 
  if ( ss->verbose >= 2 ) 
    printf(" : %g\n",f);
  else if ( ss->verbose >= 1 ) 
    printf("logprob5 : %g  loglike : %g\n",f,  ss->loglike );
  return(-f); 
}
void grad_logprob5(double *x, double *gr, void *arg )
{
  int i , g,t1,t2 ;
  double /* f=0.0 ,*/ w, lambda1, lambda2 ,s1,s2 , eps1, eps2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  int T = ss->T ; 
  double *ao = x , *ad = x + T ; 
  double *ho = x + 2*ss->T , *hd = x + 3*ss->T ; 
  double *gao = gr , *gad = gr + T ; 
  double *gho = gr + 2*ss->T , *ghd = gr + 3*ss->T ; 
  double *glogc =  gr + 4*ss->T ; 
  double ssc , llc , lc , lc1 , lc2 , c , ratio , tmp ; 
  double *mu = ss->mu ; 
  double *alpha = ss->alpha ; 
  double *logc = x + 4*ss->T ; 
  double time, lt ; 

  if ( ss->verbose >= 2 ) {
    printf ( "gradient at " ) ;
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , x[i] ) ;
  }
  for ( i = 1 ; i <= ss->k ; i++ ) gr[i] = ( x[i] - mu[i] ) * alpha[i] ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else w = 1.0 ; 
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */
    lc1 = logc[t1] ;     lc2 = logc[t2] ; lc = lc1+lc2 ;  c = exp(lc) ;
    ssc = s1 + s2 + c ;    llc = lambda1 + lambda2 + c ;  ratio = ssc/llc ;
    eps1 = ( s1 - (lambda1*ratio) ) * w ; 
    eps2 = ( s2 - (lambda2*ratio) ) * w ; 
    gho[t1] -= eps1 ;   /* ao derivative */ /* minus because f -> -f */
    gao[t2] -= eps2 ; 
    gad[t2] -= - eps1 ; /* ad derivative */ /* minus because f -> -f */
    ghd[t1] -= - eps2 ; 
    tmp = c * ( digamma_dif ( ssc , c , ss->digam_exact  ) + lc - log(llc) + 1.0 - ratio ) ;
    glogc[t1] -= tmp ;     
    glogc[t2] -= tmp ; 
  }
  if ( ss->verbose >= 2 ) {
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , gr[i] ) ;
    printf ( "\n" ) ;
  }
}

double logprob6(double *x,   void *arg )
{ /* log probability assuming fixed duration game, poisson scoring rates 
   and separable variability in those rates */
  int i , g,t1,t2 ;
  double f=0.0 , w, s1,s2 , lambda1 , lambda2 , time , lt , ll1 , ll2  ;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  double *ao = x , *ad = x + ss->T ; 
  double *ho = x + 2*ss->T , *hd = x + 3*ss->T ; 
  double *logco = x + 4*ss->T ; 
  double *logcd = x + 5*ss->T ; 
  double thisf , sc1 , lamc1 , sc2 , lamc2 , stot, logc1 , logc2 , c1 , c2 ; 
  double *mu = ss->mu ; 
  double *alpha = ss->alpha ; 
 
  if ( ss->verbose >= 2 ) 
    {  printf ( "log prob. at " ) ;
       for ( i = 1 ; i <= ss->k ; i++ ) printf( "%g " , x[i] ) ;
     } 
  ss->logprior = 0.0 ; 
  for ( i = 1 ; i <= ss->k ; i++ )
    ss->logprior -= ((x[i]-mu[i])*(x[i]-mu[i]) ) * alpha[i] * 0.5 ;
  /* run through games */
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ; /* 1 is at HOME */
    t2 = s[g][3] ; s2 = (double) s[g][4] ; stot = s1 + s2 ; 
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */
    ll1 = lt + ( ho[t1] - ad[t2] ) ; 
    ll2 = lt + ( ao[t2] - hd[t1] ) ; 

    logc1 = logco[t1] + logcd[t2] ; /* consistency of score 1 */
    logc2 = logco[t2] + logcd[t1] ;
    c1 = exp(logc1) ;	    c2 = exp(logc2) ;
    sc1 = s1 + c1 ; 	    sc2 = s2 + c2 ; 
    lamc1 = lambda1 + c1 ; 
    lamc2 = lambda2 + c2 ; 

    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else  w = 1.0 ; /* I have not checked out weightings for type 5 */
    thisf  = s1 * ll1  ; 
    thisf += s2 * ll2  ; 
    thisf += gammln (sc1) - gammln(c1) + c1 * logc1 - sc1 * log(lamc1) ; 
    thisf += gammln (sc2) - gammln(c2) + c2 * logc2 - sc2 * log(lamc2) ; 
    f += w * thisf ; 
  }
  ss->loglike = f ; 
  f += ss->logprior ; 
  if ( ss->verbose >= 2 ) 
    printf(" : %g\n",f);
  else if ( ss->verbose >= 1 ) 
    printf("lp6: %g ll: %g\n" , f ,  ss->loglike );
  return(-f); 
}
void grad_logprob6(double *x, double *gr, void *arg )
{
  int i , g,t1,t2 ;
  double /* f=0.0 ,*/ w, lambda1, lambda2 ,s1,s2 , eps1, eps2;
  opt_args *ss = ( opt_args * ) arg ;
  int **s = ss->s ; 
  int T = ss->T ; 
  double *ao = x , *ad = x + T ; 
  double *ho = x + 2*ss->T , *hd = x + 3*ss->T ; 
  double *logco = x + 4*ss->T ; 
  double *logcd = x + 5*ss->T ; 
  double *gao = gr , *gad = gr + T ; 
  double *gho = gr + 2*ss->T , *ghd = gr + 3*ss->T ; 
  double *glogco =  gr + 4*ss->T ; 
  double *glogcd =  gr + 5*ss->T ; 
  double sc1 , sc2 , c1 , c2 , logc1 , logc2 , lamc1 , lamc2 , 
               ratio1 , ratio2 , tmp ; 
  double *mu = ss->mu ; 
  double *alpha = ss->alpha ; 
  double time, lt ; 

  if ( ss->verbose >= 2 ) {
    printf ( "gradient at " ) ;
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , x[i] ) ;
  }
  for ( i = 1 ; i <= ss->k ; i++ ) gr[i] = ( x[i] - mu[i] ) * alpha[i] ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ; s1 = (double) s[g][2] ;
    t2 = s[g][3] ; s2 = (double) s[g][4] ;
    if ( ss->weighted ) { 
      w = ss->w[g] ; 
    } else w = 1.0 ; 
    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lt = log(time) ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */

    logc1 = logco[t1] + logcd[t2] ; /* consistency of score 1 */
    logc2 = logco[t2] + logcd[t1] ;
    c1 = exp(logc1) ;	    c2 = exp(logc2) ;
    sc1 = s1 + c1 ; 	    sc2 = s2 + c2 ; 
    lamc1 = lambda1 + c1 ; 
    lamc2 = lambda2 + c2 ; 

    ratio1 = sc1 / lamc1 ; 
    ratio2 = sc2 / lamc2 ; 
    eps1 = ( s1 - (lambda1*ratio1) ) * w ; 
    eps2 = ( s2 - (lambda2*ratio2) ) * w ; 
    gho[t1] -= eps1 ;   /* ao derivative */ /* minus because f -> -f */
    gao[t2] -= eps2 ; 
    gad[t2] -= - eps1 ; /* ad derivative */ /* minus because f -> -f */
    ghd[t1] -= - eps2 ; 
    tmp = c1 * ( digamma_dif ( sc1 , c1, ss->digam_exact ) + logc1 - log(lamc1) + 1.0 - ratio1 ) ;
    glogco[t1] -= tmp ;     
    glogcd[t2] -= tmp ; 
    tmp = c2 * ( digamma_dif ( sc2 , c2 , ss->digam_exact ) + logc2 - log(lamc2) + 1.0 - ratio2 ) ;
    glogco[t2] -= tmp ;     
    glogcd[t1] -= tmp ; 
  }
  if ( ss->verbose >= 2 ) {
    for ( i = 1 ; i <= ss->k ; i++ ) printf ( "%g " , gr[i] ) ;
    printf ( "\n" ) ;
  }
}
void find_A6 ( opt_args *ss ) 
{
  int i , g,t1,t2 , j ;
  double  w, lambda1, lambda2 ,s1,s2  ;
  int **s = ss->s ; 
  double time ; 
  int T = ss->T ; 
  double *ao = ss->a  ;   double *ad = ss->a + T  ; 
  double *ho = ss->a + 2*T , *hd = ss->a + 3*T ; 
  double *logco = ss->a + 4*ss->T ; 
  double *logcd = ss->a + 5*ss->T ; 
  double **A = ss->A.m  ; 
  double d2fdl2_1 ,    d2fdl2_2 ,    d2fdldc_1,    d2fdldc_2 , d2fdc2_1 , d2fdc2_2 ; 
  double sc1 , sc2 , c1 , c2 , logc1 , logc2 , lamc1 , lamc2 , 
               ratio1 , ratio2 , tmp ; 
  double *alpha = ss->alpha ; 
  int type = ss->type ; 

  for ( i = 1 ; i <= ss->T ; i++ ) { /* arbitrary 
					curvature in `n' directions */
    for ( j = 1 ; j <= ss->T ; j++ ) {
      A[i][j] = ss->curvature_n ;
      if ( type >= 2 ) A[i+T][j+T] = ss->curvature_n ;
      if ( type >= 4 ) {
	A[i+2*T][j+2*T] = ss->curvature_n ;
	A[i+3*T][j+3*T] = ss->curvature_n ;
      } /* no need for this curvature in the other dirns */
    }
  }
  for ( i = 1 ; i <= ss->k ; i++ ) 
    A[i][i] += alpha[i] ;
  for ( g = 1 ; g <= ss->G ; g ++ ) {
    t1 = s[g][1] ;  s1 = (double) s[g][2] ;
    t2 = s[g][3] ;  s2 = (double) s[g][4] ;

    if ( ss->times_known && !(ss->times_equal) ) {
      time = ss->time[g] ;
    } else time = 1.0 ; 
    lambda1 = time * exp ( ho[t1] - ad[t2] ) ; 
    lambda2 = time * exp ( ao[t2] - hd[t1] ) ; /* expected scoring rates */

    logc1 = logco[t1] + logcd[t2] ; /* consistency of score 1 */
    logc2 = logco[t2] + logcd[t1] ;
    c1 = exp(logc1) ;	    c2 = exp(logc2) ;
    sc1 = s1 + c1 ; 	    sc2 = s2 + c2 ; 
    lamc1 = lambda1 + c1 ; 
    lamc2 = lambda2 + c2 ; 

    ratio1 = sc1 / lamc1 ; 
    ratio2 = sc2 / lamc2 ; 

    if ( ss->weighted ) { 
      w = ss->w[g] ; 
      lambda1 *= w ; lambda2 *= w ; 
    } 

    d2fdl2_1 = - ratio1 * c1 * lambda1 / lamc1 ; 
    d2fdl2_2 = - ratio2 * c2 * lambda2 / lamc2 ; 
    d2fdldc_1 = c1 * lambda1 * ( s1 - lambda1 ) / ( lamc1 * lamc1 ) ; 
    d2fdldc_2 = c2 * lambda2 * ( s2 - lambda2 ) / ( lamc2 * lamc2 ) ; 
    tmp = c1 * ( digamma_dif ( sc1 , c1, ss->digam_exact ) + logc1 - log(lamc1) + 1.0 - ratio1 ) ;
    d2fdc2_1 = c1 * c1 * digamma_dif_deriv ( sc1 , c1 ) 
      + c1 * ( lambda1 * lambda1 + s1 * c1 ) / ( lamc1 * lamc1 ) 
	+ tmp ; 
    tmp = c2 * ( digamma_dif ( sc2 , c2 , ss->digam_exact ) + logc2 - log(lamc2) + 1.0 - ratio2 ) ;
    d2fdc2_2 = c2 * c2 * digamma_dif_deriv ( sc2 , c2 ) 
      + c2 * ( lambda2 * lambda2 + s2 * c2 ) / ( lamc2 * lamc2 ) 
	+ tmp ; 
    
/* lambda curvature */
/* diagonal terms */
    A[t1+2*T][t1+2*T] += - d2fdl2_1 ;      /* ao1-ao1 */
    A[t2+T][t2+T]     += - d2fdl2_1 ;  /* ad2-ad2 */
/* off-diagonal */
    A[t1+2*T][t2+T]   += d2fdl2_1 ; /* home o 1   away d 2 */
    A[t2+T][t1+2*T]   += d2fdl2_1 ;     /* o1 - d2 */
    
    A[t1+3*T][t2]     += d2fdl2_2 ; /* home  d 1   away  o 2 */
    A[t2][t1+3*T]     += d2fdl2_2 ; 
    A[t1+3*T][t1+3*T] += - d2fdl2_2 ; 
    A[t2][t2]         += - d2fdl2_2 ; 

    /* c curvature */
/* score 1 informs about co[t1],cd[t2] - o is 4, d is 5 */
    A[t1+4*T][t1+4*T] += - d2fdc2_1 ;
    A[t2+5*T][t2+5*T] += - d2fdc2_1 ;

    A[t1+4*T][t2+5*T] += - d2fdc2_1 ;
    A[t2+5*T][t1+4*T] += - d2fdc2_1 ;

/* score 2 informs about co[t2],cd[t1] - o is 4, d is 5 */
    A[t2+4*T][t2+4*T] += - d2fdc2_2 ;
    A[t1+5*T][t1+5*T] += - d2fdc2_2 ;

    A[t1+5*T][t2+4*T] += - d2fdc2_2 ;
    A[t2+4*T][t1+5*T] += - d2fdc2_2 ;

/* cross lambda-c terms     d2fdldc_1 . l1 is +o_t1 -d_t2   */
    A[t1+4*T][t1+2*T] += - d2fdldc_1 ;
    A[t1+4*T][t2+  T] +=   d2fdldc_1 ;
    A[t2+5*T][t1+2*T] += - d2fdldc_1 ;
    A[t2+5*T][t2+  T] +=   d2fdldc_1 ;

    A[t1+2*T][t1+4*T] += - d2fdldc_1 ;
    A[t2+  T][t1+4*T] +=   d2fdldc_1 ;
    A[t1+2*T][t2+5*T] += - d2fdldc_1 ;
    A[t2+  T][t2+5*T] +=   d2fdldc_1 ;

/* cross lambda-c terms     d2fdldc_2 . l2 is +o_t2 -d_t1   */
    A[t2+5*T][t2+0  ] += - d2fdldc_2 ;
    A[t2+5*T][t1+3*T] +=   d2fdldc_2 ;
    A[t1+4*T][t2+0  ] += - d2fdldc_2 ;
    A[t1+4*T][t1+3*T] +=   d2fdldc_2 ;

    A[t2+0  ][t2+5*T] += - d2fdldc_2 ;
    A[t1+3*T][t2+5*T] +=   d2fdldc_2 ;
    A[t2+0  ][t1+4*T] += - d2fdldc_2 ;
    A[t1+3*T][t1+4*T] +=   d2fdldc_2 ;

  }
}


/************************ boring stuff *********************************/

static void defaults ( opt_args *s ) 
{
#include "_var_def.c"
}

static int process_command ( int argc , char **argv , opt_args *s ) {

  int p_usage = 0 ;
  int status = 0 ;
  int cs , i ;

  if ( argc < 1 )     {
    p_usage = 1 ; 
    status -- ;
  }

#define ERROR1 fprintf ( stderr , "arg to `%s' missing\n" , argv[i] ) ; \
               status --
#define ERROR2 fprintf ( stderr , "args to `%s' missing\n" , argv[i] ) ; \
               status --
#define ERRORREG fprintf ( stderr , "regtype must be defined before `%s'\n" , argv[i] ) ; \
               status --
  for (i = 1 ; i < argc; i++)    {
    cs = 1 ;
    if ( strcmp (argv[i], "-V") == 0 )        {
      s->verbose = 1;
    }
#include "_var_clr.c"
    else {
      fprintf ( stderr , "arg `%s' not recognised\n" , argv[i] ) ; 
      p_usage = 1 ;
      status -- ;
    }
    if ( cs == 0 ) {
      fprintf ( stderr , "arg at or before `%s' has incorrect format\n" , 
	       argv[i] ) ;
      p_usage = 1 ;
      status -- ;
    }
  }
  if ( p_usage ) print_usage ( argv , stderr , s ) ;
  return ( status ) ;
}
#undef ERROR1
#undef ERROR2
#undef ERRORREG

#define DNT fprintf( fp, "\n        ")
#define NLNE  fprintf( fp, "\n")

static void print_usage ( char **argv , FILE * fp , opt_args *s )
{
  fprintf( fp, "Usage: %s ",argv[0]);
  fprintf( fp, " [optional arguments]");
  DNT;  fprintf( fp, "-V | -VV                 (verbose or very verbose)"); 
  NLNE;
#include "_var_usg.c"
  NLNE; 
  fprintf( fp, "\n");
  return ;
}
